; ==============================================================================
; file: "source/system/kernel/display.inc"
; use: to be included in "source/system/kernel.asm"
; TODO: support also other resolutions
; ==============================================================================

display_fill:

; -------------------------------------------------
; description: fills temp buffer with a solid color
; input: dword [color_1] = solid color
; output: {none}
; altered: {none}
; -------------------------------------------------

    .save_registers:
        pusha                           ; registers are preserved

    .get_temp_buffer:
        mov edi, temp_buffer            ; temp buffer starts at 0x00300000

    .reset_counter:
        mov ecx, 0                      ; counter is in ecx

    .draw_pixel:
        mov eax, dword [color_1]        ; we get color data
        mov dword [edi], eax            ; and we paste it to pixel
        add edi, 4                      ; then, we go to the next pixel

    .check_for_end:
        add ecx, 1                      ; if we have reached the end of screen
        cmp ecx, 0x307200               ; we have to return
        jne .draw_pixel

    .return:
        popa                            ; registers are restored
        ret

; ==============================================================================

display_update:

; -----------------------------------------------
; description: copies temp buffer to frame buffer
; input: {none}
; output: {none}
; altered: {none}
; -----------------------------------------------

    .prepare:
        pusha                           ; registers are preserved

    .get_frame_buffer:
        mov edi, dword [0x00007828]     ; frame buffer address is in 0x00007A28

    .get_temp_buffer:
        mov esi, temp_buffer            ; temp buffer starts at 0x00300000

    .reset_counter:
        mov ecx, 0                      ; counter is in ecx

    .check_bit_depth:
        cmp byte [0x00007819], 24       ; bit depth is in 0x00007E19
        je display_update_24            ; based on bit depth indicated here
        cmp byte [0x00007819], 32       ; we go either to the 24-bit branch
        je display_update_32            ; either to the 32-bit branch

    .return:
        popa                            ; else, we don't support this
        ret                             ; registers also are restored

; ------------------------------------------------------------------------------

display_update_24:

; ---------------------------------------------------
; description: 24 bit branch for glaux_display_update
; warning: not to be used directly
; ---------------------------------------------------

    .copy_pixel:
        mov eax, dword [esi]            ; data is moved from temp buffer
        mov dword [edi], eax            ; to frame buffer
        add esi, 4                      ; and we go to next pixel
        add edi, 3

    .check_for_end:
        add ecx, 1                      ; if we have reached the end of screen
        cmp ecx, 307200                 ; we have to return
        jne .copy_pixel                 ; screen in 640x480 has 307200 pixels

    .return:
        popa                            ; registers are restored
        ret

; ------------------------------------------------------------------------------

display_update_32:

; ---------------------------------------------------
; description: 32 bit branch for glaux_display_update
; warning: not to be used directly
; ---------------------------------------------------

    .copy_pixel:
        mov eax, dword [esi]            ; data is moved from temp buffer
        mov dword [edi], eax            ; to frame buffer
        add esi, 4                      ; and we go to next pixel
        add edi, 4

    .check_for_end:
        add ecx, 1                      ; if we have reached the end of screen
        cmp ecx, 307200                 ; we have to return
        jne .copy_pixel                 ; screen in 640x480 has 307200 pixels

    .return:
        popa                            ; registers are restored
        ret

; ==============================================================================

display_paste_str:

; -----------------------------------------------------------
; description: pastes a string to temp buffer
; input: dword [string_1] = address of string
;        dword [color_1] = background color in case of scroll
;        dword [color_2] = color of string
;        dword [pixel_horiz] = starting horiz pos
;        dword [pixel_vert] = starting vert pos
; output: dword [pixel_horiz] = ending horiz pos
;         dword [pixel_vert] = ending vert pos
; altered: dword [char_unicode]
; -----------------------------------------------------------
    
    .prepare:
        pusha                           ; registers are preserved
        mov ebx, dword [string_1]       ; string address is provided in string_1
        
    .get_char:
        mov si, word [ebx]              ; char we want to print
        mov word [char_unicode], si     ; is at that address
    
    .compare:
        cmp word [char_unicode], 0      ; if null terminator, end of string
        je .return
    
    .print:
        call display_paste_char         ; else, valid char to print
    
    .go_right:
        add ebx, 2                      ; we go to next little-endian ucs-2 char
        add dword [pixel_horiz], 8      ; we go 8 pixels right
        cmp dword [pixel_horiz], 640    ; if right border, we need to go down
        jb .get_char
    
    .go_down:
        mov dword [pixel_horiz], 0      ; we go to left border
        add dword [pixel_vert], 16      ; we go 16 pixels down
        cmp dword [pixel_vert], 480     ; if bottom border, we need to scroll
        jb .get_char
    
    .scroll:
        call display_scroll
        jmp .get_char
    
    .return:
        popa                            ; registers are restored
        ret

; ==============================================================================

display_paste_rect:

; ----------------------------------------------------
; description: pastes a rectangle to temp buffer
; input: dword [pixel_horiz] = starting horiz position
;        dword [pixel_vert] = starting vert position
;        dword [rect_width] = width of rectangle
;        dword [rect_height] = height of rectangle
;        dword [color_1] = color of rectangle
; output: {none}
; altered: {none}
; ----------------------------------------------------

    .reset_counters:
        pusha                           ; registers are preserved
        mov dword [.local_x], 0         ; we reset both the horizontal
                                        ; and the vertical counter in one access
    .check_pixel_position:
        cmp dword [pixel_horiz], 640    ; if x position over 640, out of screen
        jae .go_to_next_pixel
        cmp dword [pixel_vert], 480     ; if y position over 480, out of screen
        jae .go_to_next_pixel

    .set_buffer_position:
        mov esi, dword [pixel_vert]     ; here we are inside the screen borders
        imul esi, 2560                  ; there are 2560 bytes per row
        add esi, dword [pixel_horiz]    ; there are 4 bytes per pixel
        add esi, dword [pixel_horiz]    ; after adding 0x00300000
        add esi, dword [pixel_horiz]    ; we have the final address
        add esi, dword [pixel_horiz]
        add esi, temp_buffer

    .draw_pixel:
        mov eax, dword [color_1]        ; we copy color to pixel data
        mov dword [esi], eax

    .go_to_next_pixel:
        add dword [pixel_horiz], 1      ; we move to next pixel right
        add word [.local_x], 1          ; we increment width counter
        mov ax, word [.local_x]         ; if end of width, we go to next row
        cmp ax, word [rect_width]
        jne .check_pixel_position

    .go_to_next_row:
        mov ax, word [rect_width]       ; we go to start of width
        sub word [pixel_horiz], ax
        add dword [pixel_vert], 1       ; we move to next row
        mov word [.local_x], 0          ; we reset width counter
        add word [.local_y], 1          ; we increment height counter
        mov ax, word [.local_y]         ; if end of height, we return
        cmp ax, word [rect_height]
        jne .check_pixel_position

    .return:
        mov ax, word [rect_height]      ; we restore pixel_vert variable
        sub word [pixel_vert], ax
        popa                            ; registers are restored
        ret
    
    .local_x dw 0                       ; horizontal counter
    .local_y dw 0                       ; vertical counter

; ==============================================================================

display_paste_char:

; -----------------------------------------
; description: pastes a char to temp buffer
; input: word [char_unicode] = unicode char
;        dword [pixel_horiz] = horiz pos
;        dword [pixel_vert] = vert pos
;        dword [color_2] = color of char
; output: {none}
; altered: {none}
; -----------------------------------------

    .prepare:
        pusha                           ; registers are preserved

    .get_char_bitmap:
        movzx esi, word [char_unicode]  ; bitmap of the char is at 0x00200000
        imul esi, 16                    ; added with char_array * bitmap_length
        add esi, font_buffer

    .reset_counter:
        mov word [.local_x], 0          ; we reset both the horizontal
                                        ; and the vertical counter in one access
    .check_pixel_position:
        cmp dword [pixel_horiz], 640    ; if x position over 640, out of screen
        jae .go_to_next_pixel
        cmp dword [pixel_vert], 480     ; if y position over 480, out of screen
        jae .go_to_next_pixel

    .set_buffer_position:
        mov edi, dword [pixel_vert]     ; here we are inside the screen borders
        imul edi, 2560                  ; there are 2560 bytes per row
        add edi, dword [pixel_horiz]    ; there are 4 bytes per pixels
        add edi, dword [pixel_horiz]    ; after adding 0x00300000
        add edi, dword [pixel_horiz]    ; we have the final address
        add edi, dword [pixel_horiz]
        add edi, temp_buffer

    .eliminate_bit:
        mov cl, byte [.local_x]         ; we get the corresponding bit
        mov ah, byte [esi]              ; of current byte of bitmap
        shl ah, cl                      ; to decide if the pixel
        and ah, 0x80                    ; is opaque (1) or transparent (0)
        jz .go_to_next_pixel

    .draw_pixel:
        mov eax, dword [color_2]        ; here it is opaque, so we draw it
        mov dword [edi], eax

    .go_to_next_pixel:
        add dword [pixel_horiz], 1      ; we go to next horizontal pixel
        add byte [.local_x], 1          ; we increase horizontal counter
        cmp byte [.local_x], 8          ; char width is 8, so we must go down
        jne .check_pixel_position       ; if horizontal counter reaches 8

    .go_to_next_row:
        add esi, 1                      ; we move to next byte of bitmap
        sub dword [pixel_horiz], 8      ; we go 8 pixels left
        add dword [pixel_vert], 1       ; we go 1 pixel down
        mov byte [.local_x], 0          ; we reset horizontal counter
        add byte [.local_y], 1          ; we increment vertical counter
        cmp byte [.local_y], 16         ; char height is 16, so we must return
        jne .check_pixel_position       ; if vertical counter reaches 16

    .return:
        sub dword [pixel_vert], 16      ; we restore pixel_vert variable
        popa                            ; registers are also restored
        ret
    
    .local_x db 0                       ; horizontal counter
    .local_y db 0                       ; vertical counter

; ==============================================================================

display_scroll:

; -----------------------------------------------------
; description: scrolls the screen if it fills up
; input: dword [color_1] = background color
; output: dword [pixel_horiz] = horiz pos for next char
;         dword [pixel_vert] = vert pos for next char
; altered: dword [rect_width]
;          dword [rect_height]
; -----------------------------------------------------
    
    .prepare:
        pusha                           ; registers are preserved
    
    .set_temp_buffer_source:
        mov esi, temp_buffer + 16*640*4 ; there 16*640*4 bytes per 16 rows
                                        ; of pixels, so data will moved
                                        ; one line (16 rows) up
    .set_temp_buffer_dest:
        mov edi, temp_buffer
    
    .set_counter:
        mov ecx, 0                      ; counter is reset
    
    .copy:
        mov eax, dword [esi]            ; pixel data is moved one line up
        mov dword [edi], eax
    
    .go_to_next:
        add esi, 4                      ; we go to next pixel
        add edi, 4                      ; if we have reached 307200 pixels
        add ecx, 1                      ; we have only to clear last line
        cmp ecx, 307200
        jbe .copy
    
    .clear_last_line:
        sub dword [pixel_vert], 16      ; we just draw a rect over the last line
        mov dword [rect_width], 640
        mov dword [rect_height], 16
        call display_paste_rect
    
    .return:
        popa                            ; registers are restored
        ret
