; ==============================================================================
; file: "source/system/kernel/keybd.inc"
; use: to be included in "source/system/kernel.asm"
; ==============================================================================

keybd_get_scan:

; ------------------------------------------------------------------
; description: checks for a key
; input: {none}
; output: byte [keybd_key] = actual key pressed
;         byte [keybd_scan] = final scan code produced
;         byte [keybd_shift] = whether shift is pressed or not
; altered: eax
; TODO: check if there is valid scan code as @iansjack suggested
;       on http://forum.osdev.org/viewtopic.php?f=1&t=29205&start=14
; ------------------------------------------------------------------

    .get_scancode:
        in al, 0x60                     ; keyboard port is port 0x60
        cmp al, 0x2A                    ; left shift pressed scan code is 0x2A
        je .set_shift
        cmp al, 0x36                    ; right shift pressed scan code is 0x36
        je .set_shift
        cmp al, 0xAA                    ; left shift released scan code is 0xAA
        je .clear_shift
        cmp al, 0xB6                    ; right shift released scan code is 0xB6
        je .clear_shift
        cmp al, 0x80                    ; bit 7 set of scan code means release
        jae .set_none
        cmp al, 0                      ; if scan code 0, no key pressed/released
        je .set_none
        cmp al, byte [keybd_key]        ; if scan code equals previous scan code
        je .check_delay                 ; we can't just accept it

    .set_scancode:
        mov byte [keybd_scan], al       ; here the scan code is saved
        mov byte [keybd_key], al        ; and delay reset as it is a new key
        mov dword [.delay], 0
        jmp .return

    .check_delay:
        mov byte [keybd_scan], 0        ; here scan code is reset
        add dword [.delay], 1           ; and delay incremented
        cmp dword [.delay], 25          ; if delay has passed,
        jbe .return                     ; actual scan code is accepted
        mov byte [keybd_scan], al
        jmp .return

    .set_none:
        mov byte [keybd_scan], 0        ; scan code is cleared
        mov byte [keybd_key], 0         ; key press is also cleared
        jmp .return
        
    .set_shift:
        mov byte [keybd_shift], 1       ; shift is enabled
        jmp .return
    
    .clear_shift:
        mov byte [keybd_shift], 0       ; shift is disabled
        
    .return:
        ret

    .delay dd 0

; ==============================================================================

keybd_conv_latin:

; ------------------------------------------------------------
; description: converts a scan code to latin unicode character
; input: byte [keybd_scan] = scan code from keybd
; output: word [char_unicode] = resulting unicode char
; altered: eax, ebx
; ------------------------------------------------------------

    .save_scan_code:
        movzx ebx, byte [keybd_scan]    ; final scan code is saved

    .check_shift:
        cmp byte [keybd_shift], 0       ; if shift not pressed
        je .find_normal                 ; we search a not shifted char
    
    .find_shifted:
        mov eax, keybd_layout + 0x100   ; layout table starts at 00100100h
        jmp .save_char

    .find_normal:
        mov eax, keybd_layout           ; layout table starts at 00100000h

    .save_char:
        add eax, ebx                    ; add scancode 2 times, there are
        add eax, ebx                    ; 2 chars per entry (16-bit unicode)
        mov bx, word [eax]              ; save unicode character
        mov word [char_unicode], bx
        and dword [char_unicode], 0xFFFF

    .return:
        ret

; ==============================================================================

keybd_input_str:

; ------------------------------------------------------------
; description: inputs a unicode string using the keyboard
; input: dword [string_1] = address of string buffer
;        dword [pixel_horiz] = starting horiz pos
;        dword [pixel_vert] = starting vert pos
;        dword [color_2] = color of string
; output: dword [pixel_horiz] = ending horiz pos
;         dword [pixel_vert] = ending vert pos
; altered: eax, ebx, ecx, esi, edi
;          dword [keybd_scan]
;          dword [char_unicode]
; limitation: works nicely only with positions multiples of 16
; ------------------------------------------------------------

    .prepare:
        mov dword [.offset], 0          ; offset counter is reset

    .get_key:
        call keybd_get_scan             ; scan code is got and processed
        cmp byte [keybd_scan], 0        ; if none, we wait updating the screen
        je .update
        cmp byte [keybd_scan], 0x1C     ; if enter, we accept the string
        je .return
        cmp byte [keybd_scan], 0x0E     ; if backspace, we delete last char
        je .go_back
        call keybd_conv_latin           ; else it is a normal key
        cmp word [char_unicode], 0      ; and we need to convert it
        je .update                      ; if 0, we wait updating the screen
        call display_paste_char         ; else, we paste it to screen
    
    .save_char:
        mov esi, dword [string_1]       ; we get string buffer address
        add esi, dword [.offset]        ; we add offset to it
        mov ax, word [char_unicode]     ; and char is saved to resulting address
        mov word [esi], ax
        
    .go_right:
        add dword [.offset], 2          ; we move to next char placeholder
        add dword [pixel_horiz], 8      ; and we go 8 pixels right
        cmp dword [pixel_horiz], 640    ; except right border
        jb .update
    
    .go_down:
        mov dword [pixel_horiz], 0      ; we go to left border
        add dword [pixel_vert], 16      ; and 16 pixels down
        cmp dword [pixel_vert], 480     ; except bottom border
        jb .update
    
    .scroll:
        call display_scroll             ; we need to scroll to go down
        jmp .update
    
    .go_back:
        cmp dword [.offset], 0          ; if we are at the start of string
        je .update                      ; we can't go back
        cmp dword [pixel_horiz], 0      ; if we are at the left border,
        je .go_up                       ; we need to go to previous line
    
    .go_left:
        sub dword [pixel_horiz], 8      ; else we go 8 pixels left
        mov dword [rect_width], 8       ; and we draw over deleted char
        mov dword [rect_height], 16
        call display_paste_rect
        jmp .delete_char
    
    .go_up:
        cmp dword [pixel_vert], 0       ; if we are at the top border
        je .update                      ; we can't go more up
        mov dword [pixel_horiz], 632    ; else we go to the end of previous line
        sub dword [pixel_vert], 16      ; and we draw over deleted char
        mov dword [rect_width], 8
        mov dword [rect_height], 16
        call display_paste_rect
    
    .delete_char:
        sub dword [.offset], 2          ; we also need to delete this char
        mov esi, dword [string_1]       ; for the buffer, so we go back
        add esi, dword [.offset]        ; and overwrite it with a nice zero
        mov word [esi], 0

    .update:
        call display_update             ; we update the display so we can wait
        jmp .get_key                    ; a bit for the next char from keyboard
    
    .return:
        mov esi, dword [string_1]       ; at the end of function
        add esi, dword [.offset]        ; string must be null terminated
        mov word [esi], 0
        ret
    
    .offset dd 0
